# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`names` exposes an API to allow modifications (add, delete, modify) to 
name resolution objects for EGD devices within the Kepware Configuration API
"""

from ...utils import _url_parse_object, path_split
from ...connection import server
from ...error import KepHTTPError, KepError
from typing import Union
from .. import channel, device

NAMES_ROOT = '/name_resolution_groups/Name Resolutions/name_resolutions'

def _create_url(device_path, name = None):
    '''Creates url object for the "name resolution" branch of Kepware's project 
    tree. Used to build a part of Kepware Configuration API URL structure

    Returns the name resolution specific url when a value is passed.
    '''
    path_obj = path_split(device_path)
    device_root = channel._create_url(path_obj['channel']) + device._create_url(path_obj['device'])

    if name == None:
        return '{}/{}'.format(device_root, NAMES_ROOT)
    else:
        return '{}/{}/{}'.format(device_root, NAMES_ROOT, _url_parse_object(name))

def add_name_resolution(server: server, device_path: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"name resolution"` or multiple `"name resolution"` objects to Kepware. This allows you to 
    create a name resolution or multiple name resolutions all in one function, if desired.

    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param DATA: Dict or List of Dicts of name resolutions
    expected by Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    name resolutions added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(device_path), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_name_resolution(server: server, device_path: str, name: str) -> bool:
    '''Delete a `"name resolution"` object in Kepware.
    
    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param name: name of name resolution to delete
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(device_path, name))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_name_resolution(server: server, device_path: str, DATA: dict, *, name: str = None, force: bool = False) -> bool:
    '''Modify a `"name resolution"` object and it's properties in Kepware. If a `"name"` is not provided as an input,
    you need to identify the name resolution in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the name resolution that is to be modified.

    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param DATA: Dict of name resolution properties to be modified
    :param name: *(optional)* name of name resolution to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    name_data = server._force_update_check(force, DATA)
    if name == None:
        try:
            r = server._config_update(server.url + _create_url(device_path, name_data['common.ALLTYPES_NAME']), name_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = f'Error: No name resolution identified in DATA | Key Error: {type(DATA)}'
            raise KepError(err_msg) 
    else:
        r = server._config_update(server.url + _create_url(device_path, name), name_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_name_resolution(server: server, device_path: str, name: str = None, *, options: dict = None) -> Union[dict, list]:
    '''Returns the properties of the `"name resolution"` object or a list of all name resolutions.
    
    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param DATA: Dict of name resolution properties to be modified
    :param name: *(optional)* name of name resolution to retrieve. If not defined, will get all name resolutions
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate when getting a list of profiles. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`. Only used when `"name"` is not defined.
    
    :return: Dict of the name resolution properties or List of Dicts for all name resolutions and their properties 
    
    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    if name == None:
        r = server._config_get(f'{server.url}{_create_url(device_path)}', params= options)
    else:
        r = server._config_get(f'{server.url}{_create_url(device_path, name)}')
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`udd` module provides support for Universal Device driver specific objects 
within the Kepware Configuration API
"""

from . import profile


from ..utils import _url_parse_object
from ..connection import server
from ..error import KepHTTPError
from enum import Enum

r"""`common` contains common internal functions and constants used by the 
`ua_gateway` module.
"""

CERT_TRUST_KEY = 'ua_gateway.UA_CERTIFICATE_TRUST_STATUS'

# URL Constants for UA Gateway Module

UA_GATEWAY_ROOT = '/project/_ua_gateway'
CERT_ROOT = f'{UA_GATEWAY_ROOT}/certificates'
CLIENT_ROOT = f'{UA_GATEWAY_ROOT}/ua_client_interfaces/Client Interface'
CONN_ROOT = f'{CLIENT_ROOT}/ua_client_connections'
CLIENT_CERT_ROOT = f'{CLIENT_ROOT}/certificates'
SERVER_ROOT = f'{UA_GATEWAY_ROOT}/ua_server_interfaces/Server Interface'
ENDPOINT_ROOT = f'{SERVER_ROOT}/ua_server_endpoints'
SERVER_CERT_ROOT = f'{SERVER_ROOT}/certificates'


class _INTER_TYPE(Enum):
    SERVER = 0
    CLIENT = 1
    CERTS = 2

INSTANCE_CERTIFICATE = "Instance Certificate"

def _create_url_cert(interface, certificate = None):
    '''Creates url object for the "certificate" branch of Kepware's UA Gateway. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the UA Gateway client interfaces specific certificate url when a value is passed as the certificate name.
    '''
    if interface == _INTER_TYPE.SERVER:
        if certificate == None:
            return SERVER_CERT_ROOT
        else:
            return f'{SERVER_CERT_ROOT}/{_url_parse_object(certificate)}'
    elif interface == _INTER_TYPE.CLIENT:
        if certificate == None:
            return CLIENT_CERT_ROOT
        else:
            return f'{CLIENT_CERT_ROOT}/{_url_parse_object(certificate)}'
    else:
        if certificate == None:
            return CERT_ROOT
        else:
            return '{}/{}'.format(CERT_ROOT,_url_parse_object(certificate))

def _change_cert_trust(server: server, inter_type, certificate: str, trust: bool):
    DATA = {
        CERT_TRUST_KEY: int(trust)
    }

    cert_data = server._force_update_check(True, DATA)
    r = server._config_update(server.url + _create_url_cert(inter_type, certificate), cert_data)
    if r.code == 200: return True 
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    
def _delete_cert_truststore(server: server, inter_type, certificate: str):
    r = server._config_del(server.url + _create_url_cert(inter_type, certificate))
    if r.code == 200: return True 
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def _create_url_server(ua_server_endpoint = None):
    '''Creates url object for the "ua_server_endpoints" branch of Kepware's UA Gateway. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the UA Gateway client connections specific url when a value is passed as the ua client interface name.
    '''
    if ua_server_endpoint == None:
        return ENDPOINT_ROOT
    else:
        return f'{ENDPOINT_ROOT}/{_url_parse_object(ua_server_endpoint)}'
    
def _create_url_client(ua_client_connection = None):
    '''Creates url object for the "ua_client_connections" branch of Kepware's UA Gateway. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the UA Gateway client connections specific url when a value is passed as the ua client interface name.
    '''
    if ua_client_connection == None:
        return CONN_ROOT
    else:
        return f'{CONN_ROOT}/{_url_parse_object(ua_client_connection)}'


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`tag` exposes an API to allow modifications (add, delete, modify) to 
tag and tag group objects within the Kepware Configuration API
"""

from ..connection import server
from ..error import KepError, KepHTTPError
from ..utils import _url_parse_object, path_split
from typing import Union
from . import channel, device
import inspect

TAGS_ROOT = '/tags'
TAG_GRP_ROOT = '/tag_groups'

def _create_tags_url(tag = None):
    '''Creates url object for the "tags" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the tag specific url when a value is passed as the tag name.
    '''
    if tag == None:
        return TAGS_ROOT
    else: 
        return '{}/{}'.format(TAGS_ROOT, _url_parse_object(tag))

def _create_tag_groups_url(tag_group = None):
    '''Creates url object for the "tag_group" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the tag group specific url when a value is passed as the tag group name.
    '''
    if tag_group == None:
        return TAG_GRP_ROOT
    else: 
        return '{}/{}'.format(TAG_GRP_ROOT,_url_parse_object(tag_group))

def add_tag(server: server, tag_path: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add `"tag"` or multiple `"tag"` objects to a specific path in Kepware. 
    Can be used to pass a list of tags to be added at one path location.

    :param server: instance of the `server` class
    :param device_path: path identifying where to add tag(s). Standard Kepware address decimal 
    notation string that tags exists such as "channel1.device1.tag_group1" or "channel1.device1"
    :param DATA: Dict or List of Dicts of the tag(s) to add

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    tags added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tags_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_add(url, DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def add_tag_group(server: server, tag_group_path: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add `"tag_group"` or multiple `"tag_group"` objects to a specific path in Kepware. 
    Can be used to pass a list of tag_groups and children (tags or tag groups) to be added at one 
    path location.

    :param server: instance of the `server` class
    :param tag_group_path: path identifying where to add tag group(s). Standard Kepware address decimal 
    notation string that tag groups exists such as "channel1.device1.tag_group1" or "channel1.device1"
    :param DATA: Dict or List of Dicts of the tag group(s) to add and it's children (tags or tag groups)

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    tag groups added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tag_groups_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_add(url, DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def add_all_tags(server: server, ch_dev_path: str, DATA: dict) -> Union[bool, list]:
    '''Add `"tag"` and `"tag group"` objects to a device in Kepware. To be used to 
    pass a list of tags, tag groups and/or children of tag groups (tags and tag 
    groups) to be added at once. See example below for required structure with 
    "tags" and "tag_groups" as keys:

    Example DATA:

        {
            'tags': [tag1_dict, tag2_dict,...],
            'tag_groups':[
                {
                    tag_group1_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                }, 
                {
                    tag_group2_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                },...]
        }
    
    :param server: instance of the `server` class
    :param ch_dev_path: device path identifying where to add tags and tag groups. Standard Kepware address decimal 
    notation string that tag groups exists such as "channel1.device1"
    :param DATA: Dict of the tags and tag groups to add and it's children (tags or tag groups). 

    
    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: List [tag failure list, tag group failure list] - If a "HTTP 207 - Multi-Status" is received from 
    Kepware for either tags or tag groups, a list of dict error responses for all tags and/or tag groups added that failed.
    :return: False - If tags or tag groups are not found in DATA

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    tags_result = False
    tag_groups_result = False

    # check to see if there are dict entries for tags or tag groups
    if ('tag_groups' not in DATA) and ('tags' not in DATA):
        return False

    if 'tags' in DATA:
        tags_result = add_tag(server, ch_dev_path, DATA['tags'])
    if 'tag_groups' in DATA:
        #Add all Tag Groups
        tag_groups_result = add_tag_group(server, ch_dev_path, DATA['tag_groups'])
    
    # build results return from both calls
    if tags_result == True and tag_groups_result == True:
        return True
    elif tags_result == True:
        return [[], tag_groups_result]
    elif tag_groups_result == True:
        return [tags_result, []]
    else:
        # mixed results from both tags and tag groups
        return [tags_result, tag_groups_result]

def modify_tag(server: server, full_tag_path: str, DATA: dict, force: bool = False) -> bool:
    '''Modify a `"tag"` object and it's properties in Kepware.

    :param server: instance of the `server` class
    :param full_tag_path: path identifying location and tag to modify. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"
    :param DATA: Dict of the `tag` properties to be modified
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    tag_data = server._force_update_check(force, DATA)

    path_obj = path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for x in range(0, len(path_obj['tag_path'])-1):
            url += _create_tag_groups_url(tag_group=path_obj['tag_path'][x])
        url += _create_tags_url(tag=path_obj['tag_path'][len(path_obj['tag_path'])-1])
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_update(url, tag_data)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_tag_group(server: server, tag_group_path: str, DATA: dict, force: bool = False) -> bool:
    '''Modify a `"tag group"` object and it's properties in Kepware.

    :param server: instance of the `server` class
    :param tag_group_path: path identifying location and tag group to modify. Standard Kepware address decimal 
    notation string that tag groups exists such as "channel1.device1.tag_group1"
    :param DATA: Dict of the `tag group` properties to be modified
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    tag_group_data = server._force_update_check(force, DATA)

    path_obj = path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for tg in path_obj['tag_path']:
            url += _create_tag_groups_url(tag_group=tg)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_update(url, tag_group_data)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_tag(server: server, full_tag_path: str) -> bool:
    '''Delete `"tag"` object at a specific path in Kepware.

    :param server: instance of the `server` class
    :param full_tag_path: path identifying location and tag to delete. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for x in range(0, len(path_obj['tag_path'])-1):
            url += _create_tag_groups_url(tag_group=path_obj['tag_path'][x])
        url += _create_tags_url(tag=path_obj['tag_path'][len(path_obj['tag_path'])-1])
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_del(url)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_tag_group(server: server, tag_group_path: str) -> bool:
    '''Delete `"tag group"` object at a specific path in Kepware.

    :param server: instance of the `server` class
    :param tag_group_path: path identifying location and tag group to delete. Standard Kepware address decimal 
    notation string that tag groups exists such as "channel1.device1.tag_group1"

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for tg in path_obj['tag_path']:
            url += _create_tag_groups_url(tag_group=tg)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as err:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(err))
        raise KepError(err_msg)
    r = server._config_del(url)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_tag(server: server, full_tag_path: str) -> dict:
    '''Returns the properties of the `"tag"` object at a specific path in Kepware. 

    :param server: instance of the `server` class
    :param full_tag_path: path identifying location and tag to delete. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"

    :return: Dict of data for the tag requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for x in range(0, len(path_obj['tag_path'])-1):
            url += _create_tag_groups_url(tag_group=path_obj['tag_path'][x])
        url += _create_tags_url(tag=path_obj['tag_path'][len(path_obj['tag_path'])-1])
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url)
    return r.payload

def get_all_tags(server: server, full_tag_path: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"tag"` object at a specific path in Kepware. 

    :param server: instance of the `server` class
    :param full_tag_path: path identifying location to retreive tag list. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of tags. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`
    
    :return: List of data for all tags

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(full_tag_path)
    try:
        url = f"{server.url}{channel._create_url(path_obj['channel'])}{device._create_url(path_obj['device'])}"
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tags_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url, params= options)
    return r.payload

def get_tag_group(server: server, tag_group_path: str) -> dict:
    '''Returns the properties of the "tag group" object at a specific 
    path in Kepware. Returned object is JSON.

    :param server: instance of the `server` class
    :param tag_group_path: path identifying location and tag group to retrieve properties. Standard Kepware address decimal 
    notation string that tag groups exists such as "channel1.device1.tag_group1"

    :return: Dict of data for the tag group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for tg in path_obj['tag_path']:
            url += _create_tag_groups_url(tag_group=tg)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url)
    return r.payload

def get_all_tag_groups(server:server, tag_group_path: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"tag group"` objects at a specific 
    path in Kepware.

    :param server: instance of the `server` class
    :param tag_group_path: path identifying location to retrieve tag group list and properties. Standard Kepware address decimal 
    notation string that tag groups exists such as "channel1.device1.tag_group1"
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of devices. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`

    :return: List of data for all tag groups

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = path_split(tag_group_path)
    try:
        url = f"{server.url}{channel._create_url(path_obj['channel'])}{device._create_url(path_obj['device'])}"
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tag_groups_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url, params= options)
    return r.payload

def get_full_tag_structure(server: server, path: str, *, recursive: bool = False, options: dict = None) -> dict:
    '''Returns the properties of all `"tag"` and `"tag group"` objects at a specific 
    path in Kepware. Returned object is a dict of tag list and tag group list.

    Example:

        {
            'tags': [tag1_dict, tag2_dict,...],
            'tag_groups':[tag_group1_dict, tag_group2_dict,...]
        } 

    If `recursive` is TRUE, then the call will iterate through all tag groups and get the tags and 
    tag groups of all tag group children.This would be the equivilant of asking for all tags and tag groups
    that exist below the `"path"` location. The returned object would look like below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Example with Recursive True:

        {
            'tags': [tag1_dict, tag2_dict,...],
            'tag_groups':[
                {
                    tag_group1_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                }, 
                {
                    tag_group2_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                },...]
        } 

    :param server: instance of the `server` class
    :param path: path identifying location to retreive the tag structure. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1" and must container at least the channel and device.
    :param recursive: *(optional)* If True, returns structures within the tag groups found and all of their 
    children. (default= False)
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of tags and tag groups. 
    Options are 'filter', 'sortOrder', and 'sortProperty' only.

    :return: Dict of data for the tag structure requested at "path" location

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = {}
        
    # Remove pagination options, if present. Not useful with this method since it is designed to get the full tree of items.
    if options is not None:
        remove_list = ['pageNumber','pageSize']
        [options.pop(x) for x in remove_list]
    
    r['tags'] = get_all_tags(server, path, options= options)
    r['tag_groups'] = get_all_tag_groups(server, path, options= options)
    if recursive:
        for group in r['tag_groups']:
            res = get_full_tag_structure(server, path + '.' + group['common.ALLTYPES_NAME'], recursive= recursive, options= options)
            group.update(res)
    return r



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`datalogger` module provides support for Kepware's Datalogger plug-in specific objects 
within the Kepware Configuration API
"""
from . import log_group, log_items, triggers, mapping




# -------------------------------------------------------------------------
# Copyright (c) 2023 PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`certificates` exposes an API to allow read access to 
UA Gateway plug-in instance certificate objects within the Kepware Configuration API
"""

from ..connection import server
from ..error import KepHTTPError
from ..ua_gateway.common import _INTER_TYPE, _create_url_cert, INSTANCE_CERTIFICATE

def get_instance_certificate(server: server) -> dict:
    '''Returns the properties of the UAG instance certificate object in the UAG certificate store. 
    These are UAG instance certificates that are used by UAG for trust purposes in the UA security model.
    
    :param server: instance of the `server` class
    
    :return: Dict of properties for the certificate requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_cert(_INTER_TYPE.CERTS, INSTANCE_CERTIFICATE))
    return r.payload

# def get_all_certificates(server: server,  *, options: dict = None) -> list:
#     TODO: Implement if/when multiple instance certificates can be configured.
#     '''Returns list of all UAG instance certificate objects and their properties in the UAG certificate store. 
#     These are UAG instance certificates that are used by UAG for trust purposes in the UA security model.
    
#     :param server: instance of the `server` class
#     :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of certificates. Options are `filter`, 
#         `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`
    
#     :return: List of data for all certificates in Kepware server UAG server endpoint certificate store

#     :raises KepHTTPError: If urllib provides an HTTPError
#     :raises KepURLError: If urllib provides an URLError
#     '''
#     r = server._config_get(server.url + _create_url_cert(_INTER_TYPE.CERTS), params= options)
#     return r.payload

def reissue_self_signed_instance_certificate(server: server) -> bool:
    '''Deletes and reissues a self-signed UAG instance certificate object in the UAG certificate store. 
    This is the UAG instance certificate that are used by UAG for trust purposes in the UA security model.
    
    :param server: instance of the `server` class
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url_cert(_INTER_TYPE.CERTS, INSTANCE_CERTIFICATE))
    if r.code == 200: return True 
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`connectivity` module provides functionality to manage Kepware driver configuration 
available through the Kepware Configuration API. This includes channels, devices, 
tags, tag groups and driver specific objects. 

Driver specific object support, if available, through their own modules. Currently
the GE Ethernet Global Data and Universal Device Drivers have driver specific API
support in the SDK.
"""
from . import channel, device, tag, egd, udd


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
# **Kepware Configuration (kepconfig) API Library**
# --

# This is a package to help create Python applications to conduct operations with the Kepware Configuration API. 
# This package is designed to work with all versions of Kepware that support the Configuration API including 
# Thingworx Kepware Server (TKS), Thingworx Kepware Edge (TKE) and KEPServerEX (KEP).

# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.

r"""
.. include:: ../README.md

"""
__version__ = "1.4.0"
from . import connection, error




# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`adv_tags` module provides support for Kepware's Advanced Tags plug-in 
specific objects within the Kepware Configuration API
"""

from . import adv_tag_group, average_tags, derived_tags, complex_tags, cumulative_tags, min_tags, max_tags, link_tags
ADV_TAGS_ROOT = '/project/_advancedtags'

def _adv_tag_path_split(path: str, *, isItem=False) -> dict:
    '''Used to split the standard Kepware address decimal notation into a dict that contains the 
    advanced tag path components.

    :param path: standard Kepware address in decimal notation ("_advancedtags.tg1.tg2.tg3")
    :return: dict that contains the "adv_tag_root" and "tag_path"
    :rtype: dict

    Ex: path = "_advancedtags.tg1.tg2.tg3"

    return = {'adv_tag_root': '_advancedtags', 'tag_path': ['tg1','tg2','tg3']}

    Ex: path = "_advancedtags.ch1.dev1"

    return = {'adv_tag_root': '_advancedtags', 'tag_path': ['ch1','dev1']}
    '''
    path_list = path.split('.', 2)
    path_obj = {}
    for x in range(0, len(path_list)):
        if x == 0:
            path_obj['adv_tag_root'] = path_list[0]
        elif x == 1:
            if isItem:
                path_obj['tag_path'] = path_list[1:-1]
                path_obj['item'] = path_list[-1]
            else:
                path_obj['tag_path'] = path_list[1:]
    return path_obj

def _create_adv_tags_base_url(base_url, path_obj):
    '''Creates url object for the "path_obj" which provides the adv tags tag group structure of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the advanced tag group specific url when a value is passed as the tag group name.
    '''
    url = base_url + ADV_TAGS_ROOT
    url += adv_tag_group._create_adv_tags_group_url(path_obj)

    return url


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# Note: The code within this file was created in total or in part
#  with the use of AI tools.

r"""`link_tags` exposes an API to allow modifications (add, delete, modify) to 
link tag objects within the Kepware Configuration API
"""

from ..connection import server
from ..error import KepError, KepHTTPError
from ..utils import _url_parse_object
from typing import Union
from .. import adv_tags

LINK_TAGS_ROOT = '/link_tags'

def _get_link_tags_url(tag: str = None) -> str:
    '''Creates url object for the "link_tags" branch of Kepware's project tree.
    
    Returns the link tag specific url when a value is passed as the tag name.
    '''
    if tag is None:
        return LINK_TAGS_ROOT
    else:
        return f'{LINK_TAGS_ROOT}/{_url_parse_object(tag)}'

def add_link_tag(server: server, adv_tag_group_path: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add `"link_tag"` or multiple `"link_tag"` objects to a specific path in Kepware.
    Can be used to pass a list of link tags to be added at one path location.

    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying where to add link tag(s). Standard Kepware address decimal 
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"
    :param DATA: Dict or List of Dicts of the link tag(s) to add

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    link tags added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=False)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _get_link_tags_url()

    r = server._config_add(url, DATA)
    if r.code == 201:
        return True
    elif r.code == 207:
        errors = [item for item in r.payload if item['code'] != 201]
        return errors
    else:
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_link_tag(server: server, link_tag_path: str, DATA: dict, force: bool = False) -> bool:
    '''Modify a `"link_tag"` object and its properties in Kepware.

    :param server: instance of the `server` class
    :param link_tag_path: path identifying location and link tag to modify. Standard Kepware address decimal 
    notation string including the link tag such as "_advancedtags.AdvTagGroup1.LinkTag1"
    :param DATA: Dict of the `link_tag` properties to be modified
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    link_tag_data = server._force_update_check(force, DATA)
    path_obj = adv_tags._adv_tag_path_split(link_tag_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _get_link_tags_url(path_obj['item'])

    r = server._config_update(url, link_tag_data)
    if r.code == 200:
        return True
    else:
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_link_tag(server: server, link_tag_path: str) -> bool:
    '''Delete `"link_tag"` object at a specific path in Kepware.

    :param server: instance of the `server` class
    :param link_tag_path: path identifying location and link tag to delete. Standard Kepware address decimal 
    notation string including the link tag such as "_advancedtags.AdvTagGroup1.LinkTag1"

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(link_tag_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _get_link_tags_url(path_obj['item'])

    r = server._config_del(url)
    if r.code == 200:
        return True
    else:
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_link_tag(server: server, link_tag_path: str) -> dict:
    '''Returns the properties of the `"link_tag"` object at a specific path in Kepware.

    :param server: instance of the `server` class
    :param link_tag_path: path identifying location and link tag to retrieve. Standard Kepware address decimal 
    notation string including the link tag such as "_advancedtags.AdvTagGroup1.LinkTag1"

    :return: Dict of data for the link tag requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(link_tag_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _get_link_tags_url(path_obj['item'])

    r = server._config_get(url)
    return r.payload

def get_all_link_tags(server: server, adv_tag_group_path: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"link_tag"` objects at a specific path in Kepware.

    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying location to retrieve link tag list. Standard Kepware address decimal
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"
    :param options: *(optional)* Dict of parameters to filter, sort or paginate the list of link tags. Options are `filter`,
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`

    :return: List of data for all link tags

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=False)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _get_link_tags_url()

    r = server._config_get(url, params=options)
    return r.payload